@@ -15,19 +15,21 @@ use warnings;
 use LaTeXML::Global;
 use LaTeXML::Common::Object;
 use LaTeXML::Common::Error;
+use LaTeXML::Core::Box;
+use LaTeXML::Core::Token;
+use LaTeXML::Core::Tokens;
 use base qw(LaTeXML::Core::Definition);
 
 # Known traits:
 #    isPrefix : whether this primitive is a TeX prefix, \global, etc.
 sub new {
   my ($class, $cs, $parameters, $replacement, %traits) = @_;
-  # Could conceivably have $replacement being a List or Box?
-  my $source = $STATE->getStomach->getGullet->getMouth;
-  Fatal('misdefined', $cs, $source, "Primitive replacement for '" . ToString($cs) . "' is not CODE",
+  Error('misdefined', $cs, $STATE->getStomach,
+    "Primitive replacement for '" . ToString($cs) . "' is not a string or CODE",
     "Replacement is $replacement")
-    unless ref $replacement eq 'CODE';
+    if (ref $replacement) && (ref $replacement ne 'CODE');
   return bless { cs => $cs, parameters => $parameters, replacement => $replacement,
-    locator => $source->getLocator,
+    locator => $STATE->getStomach->getGullet->getMouth->getLocator,
     %traits }, $class; }
 
 sub isPrefix {
@@ -37,29 +39,39 @@ sub isPrefix {
 sub executeBeforeDigest {
   my ($self, $stomach) = @_;
   local $LaTeXML::Core::State::UNLOCKED = 1;
-  my $pre = $$self{beforeDigest};
-  return ($pre ? map { &$_($stomach) } grep { defined $_ } @$pre : ()); }
+  my @pre = grep { defined } @{ $$self{beforeDigest} || [] };
+  return (map { &$_($stomach) } @pre); }
 
 sub executeAfterDigest {
   my ($self, $stomach, @whatever) = @_;
   local $LaTeXML::Core::State::UNLOCKED = 1;
-  my $post = $$self{afterDigest};
-  return ($post ? map { &$_($stomach, @whatever) } grep { defined $_ } @$post : ()); }
+  my @post = grep { defined } @{ $$self{afterDigest} || [] };
+  return (map { &$_($stomach, @whatever) } @post); }
 
 # Digest the primitive; this should occur in the stomach.
 sub invoke {
   my ($self, $stomach) = @_;
-  my $profiled = $STATE->lookupValue('PROFILING') && ($LaTeXML::CURRENT_TOKEN || $$self{cs});
-  my $tracing = $STATE->lookupValue('TRACINGCOMMANDS');
-  LaTeXML::Core::Definition::startProfiling($profiled, 'digest') if $profiled;
-  print STDERR '{' . $self->tracingCSName . "}\n" if $tracing;
-  my @result = ($self->executeBeforeDigest($stomach));
-  my @args   = $self->readArguments($stomach->getGullet);
-  print STDERR $self->tracingArgs(@args) . "\n" if $tracing && @args;
-  push(@result,
-    &{ $$self{replacement} }($stomach, @args),
-    $self->executeAfterDigest($stomach));
+  my $_tracing = $STATE->lookupValue('TRACING') || 0;
+  my $tracing  = ($_tracing & TRACE_COMMANDS);
+  my $profiled = ($_tracing & TRACE_PROFILE) && ($LaTeXML::CURRENT_TOKEN || $$self{cs});
 
+  LaTeXML::Core::Definition::startProfiling($profiled, 'digest') if $profiled;
+  Debug('{' . $self->tracingCSName . '}')                        if $tracing;
+  my @result = ($self->executeBeforeDigest($stomach));
+  my $parms  = $$self{parameters};
+  my @args   = ($parms ? $parms->readArguments($stomach->getGullet, $self) : ());
+  Debug($self->tracingArgs(@args)) if $tracing && @args;
+  my $replacement = $$self{replacement};
+
+  if (!ref $replacement) {
+    my $alias = $$self{alias};
+    $alias = T_CS($alias) if $alias && !ref $alias;
+    push(@result, Box($replacement, undef, undef,
+        Tokens($alias || $$self{cs}, ($parms ? $parms->revertArguments(@args) : ())),
+        (defined $replacement ? () : (isEmpty => 1)))); }
+  else {
+    push(@result, &{ $$self{replacement} }($stomach, @args)); }
+  push(@result, $self->executeAfterDigest($stomach));
   LaTeXML::Core::Definition::stopProfiling($profiled, 'digest') if $profiled;
   return @result; }
 
@@ -67,14 +79,17 @@ sub equals {
   my ($self, $other) = @_;
   return (defined $other
       && (ref $self) eq (ref $other)) && Equals($self->getParameters, $other->getParameters)
-    && Equals($$self{replacement}, $$other{replacement}); }
+    && Equals($$self{replacement},  $$other{replacement})
+    && Equals($$self{beforeDigest}, $$other{beforeDigest})
+    && Equals($$self{afterDigest},  $$other{afterDigest})
+    ; }
 
 #===============================================================================
 1;
 
 __END__
 
-=pod 
+=pod
 
 =head1 NAME
 
@@ -93,7 +108,7 @@ It extends L<LaTeXML::Core::Definition>.
 
 Primitive definitions may have lists of daemon subroutines, C<beforeDigest> and C<afterDigest>,
 that are executed before (and before the arguments are read) and after digestion.
-These should either end with C<return;>, C<()>, or return a list of digested 
+These should either end with C<return;>, C<()>, or return a list of digested
 objects (L<LaTeXML::Core::Box>, etc) that will be contributed to the current list.
 
 =head1 AUTHOR
