@@ -41,11 +41,18 @@ our @EXPORT = qw( &pathname_find &pathname_findall &pathname_kpsewhich
   &pathname_relative &pathname_absolute &pathname_to_url
   &pathname_is_absolute &pathname_is_contained
   &pathname_is_url &pathname_is_literaldata
-  &pathname_is_raw &pathname_is_reloadable
+  &pathname_is_raw
   &pathname_protocol
   &pathname_cwd &pathname_chdir &pathname_mkdir &pathname_copy
   &pathname_installation);
 
+my $ISWINDOWS;
+
+BEGIN {
+  $ISWINDOWS = $^O =~ /^(MSWin|NetWare|cygwin)/i;
+  require Win32::ShellQuote if $ISWINDOWS;
+}
+
 # NOTE: For absolute pathnames, the directory component starts with
 # whatever File::Spec considers to be the volume, or "/".
 #======================================================================
@@ -54,7 +61,6 @@ our @EXPORT = qw( &pathname_find &pathname_findall &pathname_kpsewhich
 ### my $SEP         = '/';                          # [CONSTANT]
 # Some indicators that this is not sufficient? (calls to libraries/externals???)
 # PRELIMINARY test, probably need to be even more careful
-my $ISWINDOWS   = $^O =~ /^(MSWin|NetWare|cygwin)/i;
 my $SEP         = ($ISWINDOWS ? '\\' : '/');    # [CONSTANT]
 my $KPATHSEP    = ($ISWINDOWS ? ';'  : ':');    # [CONSTANT]
 my $LITERAL_RE  = '(?:literal)(?=:)';           # [CONSTANT]
@@ -119,7 +125,8 @@ sub pathname_canonical {
   $pathname =~ s|/\./|/|g;
   # Collapse any foo/.. patterns, but not ../..
   while ($pathname =~ s|/(?!\.\./)[^/]+/\.\.(/\|$)|$1|) { }
-  $pathname =~ s|^\./||;
+  # Reduce ./foo to foo, but preserve ./
+  $pathname =~ s|^\./(.)|$1|;
   return (defined $urlprefix ? $urlprefix . $pathname : $pathname); }
 
 # Convenient extractors;
@@ -182,15 +189,6 @@ sub pathname_is_raw {
   my ($pathname) = @_;
   return ($pathname =~ /\.(tex|pool|sty|cls|clo|cnf|cfg|ldf|def|dfu)$/); }
 
-# Check whether a pathname is reloadable as a TeX definition
-sub pathname_is_reloadable {
-  my ($pathname) = @_;
-  my ($dir, $name, $type) = pathname_split($pathname);
-  # babel.sty exception:
-  # we know the same .ldf file may be reloaded with a different option,
-  # to load an adjacently defined language, so allow that.
-  return $type eq 'ldf'; }
-
 # pathname_relative($pathname,$base) => $relativepathname
 # If $pathname is an absolute, non-URL pathname,
 # return the pathname relative to $base,
@@ -306,18 +304,17 @@ sub pathname_installation {
 sub pathname_find {
   my ($pathname, %options) = @_;
   return unless $pathname;
+  $options{findfirst} = 1;
   my @paths = candidate_pathnames($pathname, %options);
-  foreach my $path (@paths) {
-    return $path if -f $path; }
-  return; }
+  return $paths[0]; }
 
 sub pathname_findall {
   my ($pathname, %options) = @_;
   return unless $pathname;
-  my @paths = candidate_pathnames($pathname, %options);
-  return grep { -f $_ } @paths; }
+  $options{findfirst} = undef;
+  return candidate_pathnames($pathname, %options); }
 
-# It's presumably cheep to concatinate all the pathnames,
+# It's presumably cheap to concatinate all the pathnames,
 # relative to the cost of testing for files,
 # and this simplifies overall.
 sub candidate_pathnames {
@@ -357,26 +354,41 @@ sub candidate_pathnames {
         push(@exts, '.' . $ext); } } }
   push(@exts, '') unless @exts;
 
-  my @paths = ();
+  my @paths        = ();
+  my @nocase_paths = ();
+  # Precompute the intended regexes to apply
+  my @regexes = ();
+  foreach my $ext (@exts) {
+    if ($name eq '*') {    # Unfortunately, we've got to test the file system NOW...
+      if ($ext eq '.*') {    # everything, except hidden files
+        push(@regexes, [qr/^[^.]/, qr/^[^.]/]); }
+      else {
+        push(@regexes, [qr/\Q$ext\E$/i,
+            qr/\Q$ext\E$/]); } }
+    elsif ($ext eq '.*') {
+      push(@regexes, [qr/^\Q$name\E\.\w+$/i,
+          qr/^\Q$name\E\.\w+$/]); }
+    else {
+      push(@regexes, [qr/^\Q$name\E\Q$ext\E$/i,
+          qr/^\Q$name\E\Q$ext\E$/]); } }
   # Now, combine; precedence to leading directories.
   foreach my $dir (@dirs) {
-    foreach my $ext (@exts) {
-      if ($name eq '*') {    # Unfortunately, we've got to test the file system NOW...
-        if ($ext eq '.*') {    # everything
-          opendir(DIR, $dir) or next;
-          push(@paths, map { pathname_concat($dir, $_) } grep { !/^\./ } readdir(DIR));
-          closedir(DIR); }
-        else {
-          opendir(DIR, $dir) or next;    # ???
-          push(@paths, map { pathname_concat($dir, $_) } grep { /\Q$ext\E$/ } readdir(DIR));
-          closedir(DIR); } }
-      elsif ($ext eq '.*') {             # Unfortunately, we've got to test the file system NOW...
-        opendir(DIR, $dir) or next;      # ???
-        push(@paths, map { pathname_concat($dir, $_) } grep { /^\Q$name\E\.\w+$/ } readdir(DIR));
-        closedir(DIR); }
-      else {
-        push(@paths, pathname_concat($dir, $name . $ext)); } } }
-  return @paths; }
+    opendir(DIR, $dir) or next;
+    my @dir_files = readdir(DIR);
+    closedir(DIR);
+    for my $local_file (@dir_files) {
+      for my $regex_pair (@regexes) {
+        my ($i_regex, $regex) = @$regex_pair;
+        if ($local_file =~ m/$i_regex/) {
+          my $full_file = pathname_concat($dir, $local_file);
+          push(@nocase_paths, $full_file);
+          if ($local_file =~ m/$regex/) {
+            # if we are only interested in the first match, return it:
+            return ($full_file) if $options{findfirst};
+            push(@paths, $full_file); } } } } }
+  # Fallback: if no strict matches were found, return any existing case-insensitive matches
+  # Defer the -f check until we are sure we need it, to keep the usual cases fast.
+  return @paths ? @paths : @nocase_paths; }
 
 #======================================================================
 our $kpsewhich      = which($ENV{LATEXML_KPSEWHICH} || 'kpsewhich');
@@ -385,6 +397,8 @@ our $kpse_toolchain = "";
 
 sub pathname_kpsewhich {
   my (@candidates) = @_;
+  # ($kpsewhich,@candidates) MUST NOT be empty to guarantee that Perl runs $kpsewhich directly
+  # rather than through a shell or cmd.exe
   return             unless $kpsewhich && @candidates;
   build_kpse_cache() unless $kpse_cache;
   foreach my $file (@candidates) {
@@ -394,7 +408,10 @@ sub pathname_kpsewhich {
   # For multiple calls, this is slower in general. But MiKTeX, eg., doesn't use texmf ls-R files!
   if ($kpse_toolchain) {
     push(@candidates, $kpse_toolchain); }
-  if ($kpsewhich && open(my $resfh, '-|', $kpsewhich, @candidates)) {
+  if ($kpsewhich && open(my $resfh, '-|',
+      # on Windows, ($kpsewhich, @candidates) is joined into a single string, which most binaries
+      # parse with CommandLineToArgvW, so the arguments must be escaped accordingly
+      $ISWINDOWS ? Win32::ShellQuote::quote_system_list($kpsewhich, @candidates) : ($kpsewhich, @candidates))) {
     my $result = <$resfh>;     # we only need the first line
     { local $/; <$resfh>; }    # discard the rest of the output
     close($resfh);             # ignore exit status (only one of @candidates exists, usually)
@@ -404,7 +421,7 @@ sub pathname_kpsewhich {
   return; }
 
 sub build_kpse_cache {
-  $kpse_cache = {};            # At least we've tried.
+  $kpse_cache = {};    # At least we've tried.
   return unless $kpsewhich;
   # This finds ALL the directories looked for for any purposes, including docs, fonts, etc
   $kpse_toolchain = "--miktex-admin" if ($ENV{"LATEXML_KPSEWHICH_MIKTEX_ADMIN"});
@@ -434,9 +451,11 @@ sub build_kpse_cache {
         if    (/^%/) { }
         elsif (/^(.*?):$/) {    # Move to a new subdirectory
           $subdir = $1;
-          $subdir =~ s|^\./||;                             # remove prefix
-          my $d = $dir . '/' . $subdir;                    # Hopefully OS safe, for comparison?
-          $skip = !grep { $d =~ /^\Q$_\E/ } @filters; }    # check if one of the TeX paths
+          $subdir =~ s|^\./||;                           # remove prefix
+          my $d = $dir . '/' . $subdir;                  # Hopefully OS safe, for comparison?
+          $skip = !grep { $d =~ /^\Q$_\E/ } @filters;    # check if one of the TeX paths
+          $skip |= ($d =~ m|-dev[$//]|) unless $LaTeXML::DEBUG{'latex-dev'};
+        }
         elsif (!$skip) {
           # Is it safe to use '/' here?
           my $sep = '/';
@@ -515,11 +534,6 @@ with a leading "literal:" protocol.
 
 Check if pathname indicates a raw TeX source or definition file.
 
-=item C<< $boole = pathname_is_reloadable($path); >>
-
-Check for pathname exceptions where the same TeX definition file
-can be meaningfully reloaded. For example, babel.sty ".ldf" files
-
 =item C<< $rel = pathname_is_contained($path,$base); >>
 
 Checks whether C<$path> is underneath the directory C<$base>; if so
